//---------------------------------------------------
// module  : f_PRIF.cc                       
// auteur  : Max Mignotte                            
// date    :                                              
// langage : C++                                      
// labo    :                                  
//---------------------------------------------------

//-----------------------------------------------
// FICHIERS INCLUS ------------------------------
//-----------------------------------------------
#include "PRIF.h"

//------------------------------------------------
// METHODES DE LA CLASSE CLUSTERSEG  -------------
//------------------------------------------------
//----------------------------------------------------------
// Constructeur de la classe ClusterSeg
//----------------------------------------------------------
ClusterSeg::ClusterSeg(int*** imgt,int lgth,int wdth)           
 {
  int i,j,k;

  //Initialisation
  length=lgth;
  width=wdth;
  SizeS=1;

  //Allocation Memoire Matrice 
  mat_img=imatrix_allocate_3d(3,length,width);
  mat_seg=fmatrix_allocate_2d(length,width);
  smat_seg=fmatrix_allocate_3d(NB_ESPACE_COUL,length,width);

  //Allocation Windows
  win_tabseg=imatrix_allocate_2d(length,width*NB_ESPACE_COUL);

  //Initialisation
  for(k=0;k<3;k++)
  for(i=0;i<length;i++) for(j=0;j<width;j++) 
     mat_img[k][i][j]=imgt[k][i][j];        
 }


//----------------------------------------------------------
// Destructeur de la classe ClusterSeg
//----------------------------------------------------------
ClusterSeg::~ClusterSeg()
{ 
  //Libere Memoire Matrice 
  if (mat_img)    free_imatrix_3d(mat_img,3);
  if (mat_seg)    free_fmatrix_2d(mat_seg);
  if (smat_seg)   free_fmatrix_3d(smat_seg,NB_ESPACE_COUL); 
   
  //Libere Windows
  if (win_tabseg)  free_imatrix_2d(win_tabseg);
}

//------------------//
//--WINDOW- N SEGS--//
//------------------//
//----------------------------------------------------------
// ComputeWindows
//----------------------------------------------------------     
void ClusterSeg::ComputeWindows(int lgth,int wdth,int nbW)
{
  int k,i,j;

  //Recal
  for(k=0;k<nbW;k++) Recal(smat_seg[k],lgth,wdth); 

 //Windows 
 for(k=0;k<nbW;k++) for(i=0;i<lgth;i++) for(j=0;j<wdth;j++)  
   win_tabseg[i][j+(k*wdth)]=(int)smat_seg[k][i][j];

 //Visu
 SizeS=nbW;
}

//----------------------------
//----<Histo><COLOR-BASE>-----
//----------------------------
//----------------------------------------------------------
//  Segmentation <Histo-COLOR-> Kmean
//       < Vraie Histo >
//
//  image[lgth][wdth] en NbClass 
//  Histo divise en <div> bin
//  SizeWindow: <Size>
// 
//  bascol::
//  -----------            
//  -0- RGB ->    [X,Y,Z] = [0,255]
//  -1- HSV ->    H=[0::360]  S=[0::1]   V=[0::255]
//  -2- YIQ -     Y=[0::1]    I=[-0.55::0.55]  Q=[-0.4::0.4]
//  -3- XYZ ->    [X,Y,Z] = [0,255]
//  -4- LAB -     L=[0::100]    A=[-128::143]  B=[-258::237]
//  -5- LUV -     L=[-16::100]  U=[-60::145]   V=[-130::117]    
//
//  -6- i123-     i1=[0::255]    i2=[-255::255]      i3=[-255::255]  
//  -7- h123-     h1=[0::510]    h2=[-255::255]      h3=[-255::255]
//  -8- YCbCr    Y=[0::255]     Cb=[-127.5::127.5]  Cr=[-127.5::127.5]
//  -9- TSL-      T=[0::1]       S=[0::1.265]        L=[0::1]
//  -10-HSL-      H=[0::3.142]   S=[0::1]            L=[0::1]    
//  -11- P1P2     P1=[-0.707::0.707]   P2=[-0.408::0.816]   
//
//  Recalage sur [0-255] avec
//
//             (C-min)*255/(max-min)
//
//----------------------------------------------------------     
//----------------------------------------------------------     
// SegmentationHistoCOLORKmean Avec Entier 
//----------------------------------------------------------     
void ClusterSeg::SegmentationHistoCOLORKmeanFast(int***  img,float** matout,int lgth,int wdth,int NbClass,int div,int SzW,int Seed,int bascol)
{
 int i,j,k,l;
 int nbpts; 
 int posr,posc;
 float divis;
 float lab;
 float col1,col2,col3;

 //Init
 int NbParam=div*div*div; 
 divis=(256.0/(float)div);  
 
 //Initialisatio memoire
 int** PtsTab=imatrix_allocate_2d(lgth*wdth,NbParam);
 float*** SquWin=fmatrix_allocate_3d(3,SzW,SzW);
 float*** ImgColorSpace=fmatrix_allocate_3d(3,lgth,wdth);
 int*  TabLabel=imatrix_allocate_1d(NbParam);
 int* VctLabel=imatrix_allocate_1d(lgth*wdth);  

 //Presentation
 printf("\n\n\n   >>> SegmentationHisto-COLOR-Kmean");
 printf("[NbCl:%d][Div:%d][Ws:%d][Col:%d][Par:%d]",NbClass,div,SzW,bascol,NbParam); 
 if (bascol==0)  printf(" >RGB");
 if (bascol==1)  printf(" >HSV");
 if (bascol==2)  printf(" >YIQ");
 if (bascol==3)  printf(" >XYZ");
 if (bascol==4)  printf(" >LAB");
 if (bascol==5)  printf(" >LUV");
 if (bascol==6)  printf(" >I1I2I3");
 if (bascol==7)  printf(" >H1H2H3");
 if (bascol==8)  printf(" >YCbCr");
 if (bascol==9)  printf(" >TSL");
 if (bascol==10) printf(" >HSL"); 
 if (bascol==11) printf(" >P1P2"); 
 fflush(stdout);
 

 //-------------------------------------------------------------------------
 //Conversion
 //-------------------------------------------------------------------------
 if (bascol==0)
 for(i=0;i<lgth;i++) for(j=0;j<wdth;j++) 
    { ImgColorSpace[0][i][j]=img[0][i][j];
      ImgColorSpace[1][i][j]=img[1][i][j];
      ImgColorSpace[2][i][j]=img[2][i][j]; }

 if (bascol==1)
 for(i=0;i<lgth;i++) for(j=0;j<wdth;j++) 
    { RGBToHSV(img[0][i][j],img[1][i][j],img[2][i][j],&col1,&col2,&col3); 
      ImgColorSpace[0][i][j]=(col1/360.0)*255.0;
      ImgColorSpace[1][i][j]=col2*255.0;
      ImgColorSpace[2][i][j]=col3; }

 if (bascol==2)
 for(i=0;i<lgth;i++) for(j=0;j<wdth;j++) 
    { RGBToYIQ(img[0][i][j],img[1][i][j],img[2][i][j],&col1,&col2,&col3); 
      ImgColorSpace[0][i][j]=col1*255.0;
      ImgColorSpace[1][i][j]=((col2+0.55)/1.10)*255.0;
      ImgColorSpace[2][i][j]=((col3+0.4)/0.8)*255.0; }

 if (bascol==3)
 for(i=0;i<lgth;i++) for(j=0;j<wdth;j++) 
    { RGBToXYZ(img[0][i][j],img[1][i][j],img[2][i][j],&col1,&col2,&col3);
      ImgColorSpace[0][i][j]=col1;
      ImgColorSpace[1][i][j]=col2;
      ImgColorSpace[2][i][j]=col3; }

 if (bascol==4)
 for(i=0;i<lgth;i++) for(j=0;j<wdth;j++) 
    { RGBToLAB(img[0][i][j],img[1][i][j],img[2][i][j],&col1,&col2,&col3);
      ImgColorSpace[0][i][j]=col1*2.55;
      ImgColorSpace[1][i][j]=((col2+128)/271.0)*255.0;
      ImgColorSpace[2][i][j]=((col3+258)/495.0)*255.0; }

 if (bascol==5)
 for(i=0;i<lgth;i++) for(j=0;j<wdth;j++) 
   {  RGBToLUV(img[0][i][j],img[1][i][j],img[2][i][j],&col1,&col2,&col3);
      ImgColorSpace[0][i][j]=((col1+16)/250.0)*255.0;
      ImgColorSpace[1][i][j]=((col2+60)/205.0)*255.0;
      ImgColorSpace[2][i][j]=((col3+130)/247.0)*255.0; }

 if (bascol==6)
 for(i=0;i<lgth;i++) for(j=0;j<wdth;j++) 
    { RGBToI1I2I3(img[0][i][j],img[1][i][j],img[2][i][j],&col1,&col2,&col3);
      ImgColorSpace[0][i][j]=col1;
      ImgColorSpace[1][i][j]=((col2+255)/2.0);
      ImgColorSpace[2][i][j]=((col3+255)/2.0);  }

 if (bascol==7)
 for(i=0;i<lgth;i++) for(j=0;j<wdth;j++) 
    { RGBToH1H2H3(img[0][i][j],img[1][i][j],img[2][i][j],&col1,&col2,&col3);
      ImgColorSpace[0][i][j]=col1/2.0; 
      ImgColorSpace[1][i][j]=((col2+255)/2.0); 
      ImgColorSpace[2][i][j]=((col3+255)/2.0); }

 if (bascol==8)
 for(i=0;i<lgth;i++) for(j=0;j<wdth;j++) 
    { RGBToYCbCr(img[0][i][j],img[1][i][j],img[2][i][j],&col1,&col2,&col3);
      ImgColorSpace[0][i][j]=col1; 
      ImgColorSpace[1][i][j]=col2+127.5; 
      ImgColorSpace[2][i][j]=col3+127.5; }

 if (bascol==9)
 for(i=0;i<lgth;i++) for(j=0;j<wdth;j++) 
    { RGBToTSL(img[0][i][j],img[1][i][j],img[2][i][j],&col1,&col2,&col3);
      ImgColorSpace[0][i][j]=col1*255;
      ImgColorSpace[1][i][j]=255*col2/1.265; 
      ImgColorSpace[2][i][j]=col3*255;  }

 if (bascol==10)
 for(i=0;i<lgth;i++) for(j=0;j<wdth;j++) 
    { RGBToHSL(img[0][i][j],img[1][i][j],img[2][i][j],&col1,&col2,&col3);
      ImgColorSpace[0][i][j]=255*col1/3.142;  
      ImgColorSpace[1][i][j]=col2*254;
      ImgColorSpace[2][i][j]=col3*255; }

 if (bascol==11)
 for(i=0;i<lgth;i++) for(j=0;j<wdth;j++) 
    { RGBToP1P2(img[0][i][j],img[1][i][j],img[2][i][j],&col1,&col2,&col3);
      ImgColorSpace[0][i][j]=(col1+0.707)*255.0/1.414; 
      ImgColorSpace[1][i][j]=(col2+0.408)*255.0/1.224; 
      ImgColorSpace[2][i][j]=0.0; }


 //Debug
 for(i=0;i<lgth;i++) for(j=0;j<wdth;j++) 
   { if (ImgColorSpace[0][i][j]>255)  ImgColorSpace[0][i][j]=255.0;
     if (ImgColorSpace[1][i][j]>255)  ImgColorSpace[1][i][j]=255.0;
     if (ImgColorSpace[2][i][j]>255)  ImgColorSpace[2][i][j]=255.0;
     if (ImgColorSpace[0][i][j]<0)    ImgColorSpace[0][i][j]=0.0;
     if (ImgColorSpace[1][i][j]<0)    ImgColorSpace[1][i][j]=0.0;
     if (ImgColorSpace[2][i][j]<0)    ImgColorSpace[2][i][j]=0.0; }
 //-------------------------------------------------------------------------- 
  

 //---------------------------------------------------------------
 //Data   
 //----------------------------------------------------------------                       
 nbpts=0;
 for(i=0;i<lgth;i++) for(j=0;j<wdth;j++)  
    {
     for(k=0;k<SzW;k++) for(l=0;l<SzW;l++) 
        { 
	 posr=i-(SzW/2)+k;
         posc=j-(SzW/2)+l;

         if (posr<0)           posr+=lgth;
         if (posr>(lgth-1))    posr-=lgth;  
         if (posc<0)           posc+=wdth;
         if (posc>(wdth-1))    posc-=wdth;
        
         SquWin[0][k][l]=ImgColorSpace[0][posr][posc];
         SquWin[1][k][l]=ImgColorSpace[1][posr][posc];
         SquWin[2][k][l]=ImgColorSpace[2][posr][posc];  
        }

     //DataStructure
     for(k=0;k<NbParam;k++)  TabLabel[k]=0;
      
     for(k=0;k<SzW;k++) for(l=0;l<SzW;l++)  
	{ lab=((int)(SquWin[0][k][l]/divis))*(div*div);
          lab+=((int)(SquWin[1][k][l]/divis))*(div);  
          lab+=((int)(SquWin[2][k][l]/divis));  
          TabLabel[(int)(lab)]++; }
    
     for(k=0;k<NbParam;k++)    PtsTab[nbpts][k]=TabLabel[k];

     nbpts++;     
    }//----------------------------------------------------------------

 //K-Mean    
 KmeanFast(PtsTab,NbClass,(lgth*wdth),NbParam,VctLabel,Seed,1);
 for(i=0;i<lgth;i++) for(j=0;j<wdth;j++)  matout[i][j]=VctLabel[(i*wdth)+j];    

 //Liberation Memoire
 if (PtsTab)         free_imatrix_2d(PtsTab);
 if (SquWin)         free_fmatrix_3d(SquWin,3);
 if (TabLabel)       free_imatrix_1d(TabLabel);
 if (VctLabel)       free_imatrix_1d(VctLabel);
 if (ImgColorSpace)  free_fmatrix_3d(ImgColorSpace,3);
}

//---------------------------------------
//---------------------------------------
//------- FUSION REGIONS ----------------
//---------------------------------------
//---------------------------------------
//----------------------------------------------------------     
// ComputeNrjLocReg   [[ Ne  modifie pas <seg> ]]
//
// << Pour FUSE-REGIONS >>
//
//----------------------------------------------------------     
float ClusterSeg::ComputeNrjLocReg(float** seg,float** bin,int cl,int clinit,float*** matpds,int NbConnections,float* distconx)
{
 int i,j;
 float Nrj;
 
 //Coloriage Region[bin](clinit) en classe (cl)
 for(i=0;i<length;i++) for(j=0;j<width;j++) if (bin[i][j]) seg[i][j]=cl; 

 //NrjPrior  
 Nrj=ComputeNrj(seg,matpds,NbConnections,distconx); 

 //Resupere <Seg>
 for(i=0;i<length;i++) for(j=0;j<width;j++) if (bin[i][j]) seg[i][j]=clinit;

 //Retour
 return Nrj;
}

//----------------------------------------------------------     
// ComputeNrj
//----------------------------------------------------------     
float ClusterSeg::ComputeNrj(float** seg,float*** matpds,int NbConnections,float* distconx)
{
 int i,j,l,m;
 float nrj;
 int dist;
 int posc,posr;

 //Nrj
 nrj=0.0;
 for(i=0;i<length;i++) for(j=0;j<width;j++)
    {
     m=(int)seg[i][j];
     for(l=0;l<(NbConnections+4);l++)
        {
         dist=(i*width)+j+(int)distconx[l];      
         dist=dist%((length*width));        
         posr=(int)(dist/width);
         posc=(int)(dist%width);
         if (posr<0)             posr+=length;
         if (posr>(length-1))    posr-=length;  
         if (posc<0)             posc+=width;
         if (posc>(width-1))     posc-=width;
         if (seg[posr][posc]!=m) nrj+=matpds[l][i][j];
         else nrj-=matpds[l][i][j];
         }
    }
 
 //Retour
 return nrj;
}

//----------------------------------------------------------
// FuseRegions
// -----------
//
// <Fusionne une region en une classe qui lui est voisine>
//       
//----------------------------------------------------------
int ClusterSeg::FuseRegionsPrior(float** seg,float*** TabSeg,int NbSegs,int sz,float regul,int* NbRegs,int NbClMin)
{
 int i,j,k,l,m,n;
 int oldclass;
 int cpt;
 int nbclassVois;
 int nbregchgt;
 float minnrj;
 int   labelmin;
 int size;
 float NrjMap;

 int NbConnections;
 int dist;
 int posc,posr;
 int nb;
 float ech;

 //Presentation
 printf("\n\n   >>> Function  > FuseRegionPrior ...[Regul:%f][NbClMin:%d]]",regul,NbClMin);
 printf("\n       ----------------------------------------------------------------");
 printf("\n");

 //-4-voisins+NbConnections 
 NbConnections=80; 

 //Allocation 
 float**  segtmp=fmatrix_allocate_2d(length,width);
 float**  segbin=fmatrix_allocate_2d(length,width);
 int*     claslb=imatrix_allocate_1d(NbCl1);
 float*   nrjtab=fmatrix_allocate_1d(NbCl1);

 float***  matpds=fmatrix_allocate_3d(NbConnections+10,length,width);
 float*    distconx=fmatrix_allocate_1d(NbConnections+10);

  //DistanceConnexion
  //-----------------------------------------------------------------
  //4-Voisins DistanceConnexion
  distconx[0]=1.0; 
  distconx[1]=width; 
  distconx[2]=-1.0; 
  distconx[3]=-width;

  //DistanceConnexion
  ech=CARRE(sz)/(float)NbConnections;
  ech=(int)(ech+0.4999);
  nb=0;
  cpt=4;
  for(k=-(sz/2);k<(sz/2);k++) for(l=-(sz/2);l<(sz/2);l++) 
    { dist=(k*width)+l;
      if (cpt>(NbConnections+4)) break;
      if (!(nb%(int)ech)) { distconx[cpt]=dist; cpt++; }     
      nb++; }

  //Info
  printf("\n  > Size[%d]  NbCnx[%d]  Ech[%.0f]  NbRegs[%d] \n",sz,cpt,ech,(*NbRegs));
  
  //Calcul Pds [Homogeneite:1 | Contour:-1]
  for(i=0;i<length;i++) for(j=0;j<width;j++) 
    for(k=0;k<NbSegs;k++)
      for(l=0;l<(NbConnections+4);l++)
	     {
               dist=(i*width)+j+(int)distconx[l]; 
               dist=dist%((length*width));              
               posr=(int)(dist/width);
               posc=(int)(dist%width);
               if (posr<0)              posr+=length;
               if (posr>(length-1))     posr-=length;  
               if (posc<0)              posc+=width;
               if (posc>(width-1))      posc-=width;
               if (TabSeg[k][i][j]!=TabSeg[k][posr][posc])  matpds[l][i][j]-=2;
             } 

  for(i=0;i<length;i++) for(j=0;j<width;j++) for(l=0;l<(NbConnections+4);l++)
    matpds[l][i][j]=(matpds[l][i][j]/NbSegs)+1;
  //-------------------------------------------------------------------

 //Verif & Init
 for(i=0;i<length;i++) for(j=0;j<width;j++) seg[i][j]=(int)seg[i][j];
 for(i=0;i<length;i++) for(j=0;j<width;j++) segbin[i][j]=0.0;
 nbregchgt=0;

 //InitMap
 NrjMap=ComputeNrj(seg,matpds,NbConnections,distconx);    
 printf("\n  >> Init >>> NrjMap=[%.0f]",NrjMap);  

 //Init 
 for(i=0;i<length;i++) for(j=0;j<width;j++) segtmp[i][j]=0;

   //BOUCLE
   //========================================================================
   for(cpt=0,i=0;i<length;i++) for(j=0;j<width;j++) if (segtmp[i][j]==0)
      {
       Paint(seg,segtmp,i,j,length,width,(int)seg[i][j]);
       ++cpt;

       //Info 
       printf("\n Region[%d:cl%d:<%d:%d>]",cpt,(int)seg[i][j],i,j);

       //Stat
       size=0;
       for(k=0;k<length;k++) for(l=0;l<width;l++) if (segtmp[k][l]==-1) size++;
       printf("|Sz=%d|",size);

       //Init
       for(k=0;k<NbCl1;k++)  claslb[k]=0;
       for(k=0;k<NbCl1;k++)  nrjtab[k]=100000000000.0;
       for(k=0;k<length;k++) for(l=0;l<width;l++) if (segtmp[k][l]==-1) segbin[k][l]=1.0;

       //ClassesVoisines
       for(k=0;k<length;k++) for(l=0;l<width;l++) if (segtmp[k][l]==-1)  
	  { for(m=-1;m<2;m++) for(n=-1;n<2;n++) 
		if (((k+m)>=0)&&((k+m)<length)&&((l+n)>=0)&&((l+n)<width))
		   claslb[(int)seg[k+m][l+n]]=1; }

       oldclass=(int)seg[i][j];

       //Info 
       nbclassVois=0;
       for(nbclassVois=0,n=0;n<NbCl1;n++) if  (claslb[n]) nbclassVois++;
       printf(">nb=%d::",nbclassVois);  
       for(m=0,n=0;n<NbCl1;n++)  if (claslb[n]) { claslb[m]=n; m++; }
       for(k=0;k<nbclassVois;k++) printf("%d-",claslb[k]);
       fflush(stdout);

       //Fusion------------------------------------
       minnrj=100000000000.0;
       labelmin=oldclass;
       
       if (nbclassVois>1)
       for(k=0;k<nbclassVois;k++)
	 {  
	   nrjtab[k]=ComputeNrjLocReg(seg,segbin,claslb[k],oldclass,matpds,NbConnections,distconx);
      
           if ((claslb[k]!=oldclass)&&((*NbRegs)>NbClMin)) nrjtab[k]-=regul;

	   printf("(%.0f)",nrjtab[k]/10.0);
	   if (nrjtab[k]<minnrj) { minnrj=nrjtab[k]; labelmin=claslb[k]; } 
         }

       printf("<Fus::%d>",labelmin);
       fflush(stdout);

       if (labelmin!=oldclass) 
          { nbregchgt++; 
	    printf("*");
            for(k=0;k<length;k++) for(l=0;l<width;l++) if (segbin[k][l]) seg[k][l]=labelmin;
            NrjMap=ComputeNrj(seg,matpds,NbConnections,distconx);    
            (*NbRegs)--;
            printf("\n  >>NrjMap=[%.0f] (NbRegs:%d)",NrjMap,(*NbRegs));  
          }
       //-------------------------------------------
             
        //Region faite
        for(k=0;k<length;k++) for(l=0;l<width;l++) if (segtmp[k][l]==-1) segtmp[k][l]=-2;
        for(k=0;k<length;k++) for(l=0;l<width;l++) segbin[k][l]=0.0;
      }//================================================================================
  
 //Liberation 
 if (segtmp)  free_fmatrix_2d(segtmp); 
 if (segbin)  free_fmatrix_2d(segbin);
 if (claslb)  free_imatrix_1d(claslb); 
 if (nrjtab)  free_fmatrix_1d(nrjtab); 

 if (matpds)   free_fmatrix_3d(matpds,4);
 if (distconx) free_fmatrix_1d(distconx);

 //Info
 printf("\n      >> NbChgtFuseRegions <[%d]>",nbregchgt);

 //Retour
 return nbregchgt;
}

//-------------------------------------------//
//-----------METHODES SEGMENTATION-----------//
//-------------------------------------------//
//----------------------------------------------------------
// Segment
//----------------------------------------------------------    
void ClusterSeg::Segmente()
{
  int k,l,m;

 //Pour Tests
 char MyBuff[100];
 strcpy(MyBuff,Extension);
 const int NB_SEGS=10;
 int NbTotSegs;
 int flag;
 int nbreg;

 //Calcul
 NbTotSegs=0;
 if (NbCl1) NbTotSegs+=NbSegs;
 if (NbCl2) NbTotSegs+=NbSegs;
 if (NbCl3) NbTotSegs+=NbSegs;
 if (NbCl4) NbTotSegs+=NbSegs;
 if (NbCl5) NbTotSegs+=NbSegs;
 if (NbCl6) NbTotSegs+=NbSegs;

 //Information
 //-----------
 printf("\n -----");
 printf("\n    > NbCl 1|2|3|4|5|6   [%d][%d][%d][%d][%d][%d]",NbCl1,NbCl2,NbCl3,NbCl4,NbCl5,NbCl6); 
 printf("\n    > NbSegs     >>>>  [%d]",NbSegs);
 printf("\n    > NbTotSegs  >>>>  [%d]",NbTotSegs);
 printf("\n -----");
 printf("\n    > DistFusion  >>>>  [%.3f]",DistBhatta);
 printf("\n    > NbClassMin  >>>>  [%d]",NbClassMin);
 printf("\n    > SizeMin     >>>>  [%d]",SizeMin);
 printf("\n -----");
 printf("\n    > NbBins [%d]",NbBins);  
 printf("\n    > Size1 | Size2  [%d][%d]",Size1,Size2); 
 printf("\n -----");
 printf("\n Extension >> [%s]",Extension);
 printf("\n -----");
 printf("\n");
 printf("\n Image: [%s]",Name_Img);
 printf("  Size:[%d - %d]",length,width);
 printf("\n");
 fflush(stdout);

 //===============================================================================
 //>ColorSegmentation 
 //===============================================================================
 if (NbCl1)
 { printf("\n\n <NbClasses [%d]>",NbCl1);
   printf("\n ---------------");
   for(k=0;k<NbSegs;k++) 
     { SegmentationHistoCOLORKmeanFast(mat_img,smat_seg[k],length,width,NbCl1,NbBins,Size1,k,k);
       for(m=0;m<5;m++) if (!FuseSmallRegions(smat_seg[k],length,width,&nbreg,SizeMin))  break; } }

 if (NbCl2)
 { printf("\n\n\n <NbClasses [%d]>",NbCl2);
   printf("\n ---------------");
   for(k=NbSegs,l=0;k<(2*NbSegs);k++,l++) 
     { SegmentationHistoCOLORKmeanFast(mat_img,smat_seg[k],length,width,NbCl2,NbBins,Size1,k,l);
       for(m=0;m<5;m++) if (!FuseSmallRegions(smat_seg[k],length,width,&nbreg,SizeMin))  break; } }

 if (NbCl3)
 { printf("\n\n\n <NbClasses [%d]>",NbCl3);
   printf("\n ---------------");
   for(k=(2*NbSegs),l=0;k<(3*NbSegs);k++,l++) 
     { SegmentationHistoCOLORKmeanFast(mat_img,smat_seg[k],length,width,NbCl3,NbBins,Size1,k,l);
       for(m=0;m<5;m++) if (!FuseSmallRegions(smat_seg[k],length,width,&nbreg,SizeMin))  break; } } 

 //Les trois suivants sont avec des histo compreneant 4^3 bins
 NbBins--;

 if (NbCl4)
 { printf("\n\n\n <NbClasses [%d]>",NbCl4);
   printf("\n ---------------");
   for(k=(3*NbSegs),l=0;k<(4*NbSegs);k++,l++) 
     { SegmentationHistoCOLORKmeanFast(mat_img,smat_seg[k],length,width,NbCl4,NbBins,Size1,k,l);
       for(m=0;m<5;m++) if (!FuseSmallRegions(smat_seg[k],length,width,&nbreg,SizeMin))  break; } } 

 if (NbCl5)
 { printf("\n\n\n <NbClasses [%d]>",NbCl5);
   printf("\n ---------------");
   for(k=(4*NbSegs),l=0;k<(5*NbSegs);k++,l++) 
     { SegmentationHistoCOLORKmeanFast(mat_img,smat_seg[k],length,width,NbCl5,NbBins,Size1,k,l);
       for(m=0;m<5;m++) if (!FuseSmallRegions(smat_seg[k],length,width,&nbreg,SizeMin))  break; } } 

 if (NbCl6)
 { printf("\n\n\n <NbClasses [%d]>",NbCl6);
   printf("\n ---------------");
   for(k=(5*NbSegs),l=0;k<(6*NbSegs);k++,l++) 
     { SegmentationHistoCOLORKmeanFast(mat_img,smat_seg[k],length,width,NbCl6,NbBins,Size1,k,l);
       for(m=0;m<5;m++) if (!FuseSmallRegions(smat_seg[k],length,width,&nbreg,SizeMin))  break; } } 

 //Retour Normal
 NbBins++;  
 //===============================================================================


  //===============================================================================
  //>Fusion
  //===============================================================================
  FuseSegmentation(smat_seg,NbTotSegs,NbCl1,mat_seg,length,width,Size2,8);  
   

  //===============================================================================
  //>Fusion Regions
  //===============================================================================
  for(;;) if (!FuseSmallRegions(mat_seg,length,width,&NbRegions,50))  break;
  printf("\n >>>>>> NbRegions [%d]",NbRegions);
  
  for(m=0;m<10;m++) { flag=FuseRegionsPrior(mat_seg,smat_seg,NbTotSegs,Size2,DistBhatta,&NbRegions,NbClassMin);
            NbRegions=ComputeNumberRegions(mat_seg,length,width);
            printf("\n >>>>>> IterBoucle[%d]   NbRegions [%d]",m,NbRegions);
            if (!flag) break; }
  //===============================================================================

  //Regions
  Recal(mat_seg,length,width);
  ConvertClassRegion(mat_seg,length,width);

  //Contour
  AddContourImg(mat_seg,mat_img,length,width,2);

  //Sauvegarde
  SaveImagePgm(Name_Img,Extension,mat_seg,length,width);
  if (flag_save) SaveImagePpm(Name_Img,Extension,mat_img,length,width); 

 //Affichage 
 if (flag_visu)  
   { ComputeWindows(length,width,NB_SEGS); }

}

